
/******************************************************************************
* MODULE     : conv_apply.cc
* DESCRIPTION: Function application
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <evaluator.hh>
#include <function.hh>
#include <xtable.hh>
namespace mmx {

void
evaluator_rep::kw_conv_apply_1 () {
  evaluate_args (KW_CONV_APPLY_2);
}

void
evaluator_rep::kw_conv_apply_2 () {
  // INPUT: (conv_apply_2, args, cont, ...)
  list l= as_list (car (rem));
  generic fun= V (car (l));
  generic arg= V (cadr (l));
  generic rtp= T (car (l)) [2];
  if (is_function (fun)) {
    function f= as_function (fun);
    cur= cadr (rem);
    rem= cons (value (f(arg), rtp), cddr (rem));
    // OUTPUT: (cont, res, ...)
  }
  else if (is_func (fun, MMX_LAMBDA, 5)) {
    list    lamb  = as_list (fun);
    list    vars  = as_list (lamb[2]);
    list    ltypes= as_list (lamb[3]);
    list    args  = cdr (as_list (arg));
    generic rtype = lamb[4];
    generic expr  = lamb[5];

    list decls;
    if ((N (vars) != N (ltypes)) || (N (vars) != N (args))) {
      error (car (trace), "Bad number of arguments");
      return;
    }
    while (!nil (vars)) {
      generic decl= list (car (vars), value (car (args), car (ltypes)));
      decls = cons (decl, decls);
      vars  = cdr (vars);
      ltypes= cdr (ltypes);
      args  = cdr (args);
    }
    decls= reverse (decls);

    gentable old_env= env;
    env= xtable (as_gentable (lamb[1]), table ());
    declare (decls);
    cur= list (MMX_TRANSTYPE, expr, value (rtype, ID_CLASS));
    rem= cons (KW_CONV_APPLY_3, old_env, cdr (rem));
    // OUTPUT: (lambda_body, conv_apply_3, old_env, cont, ...)
    return;
  }
  else if (is_func (fun, CONV_LAMBDA, 2)) {
    env (fun[1])= cadr (l);
    cur= fun[2];
    rem= cdr (rem);
    // OUTPUT: (lambda_body, cont, ...)
  }
  else {
    error (fun, "Not a function");
    return;
  }
  trace= cdr (trace);
}

void
evaluator_rep::kw_conv_apply_3 () {
  // INPUT: (conv_apply_3, lambda_ret, old_env, cont, ...)
  env= as_gentable (cadr (rem));
  cur= caddr (rem);
  rem= cons (car (rem), cdddr (rem));
  // OUTPUT: (cont, ret, ...)
  trace= cdr (trace);
}

} // namespace mmx
